@@ -7,7 +7,7 @@
 #include "pow.h"
 
 static const char*       GEOHEX3_HASH_KEY  = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";
-static const long double GEOHEX3_HASH_DEG  = (long double)M_PI / 6.0L;
+static const long double GEOHEX3_HASH_DEG  = GEOHEX3_PI / 6.0L;
 
 // inner functions
 static geohex_coordinate_t geohex_get_coordinate_by_code(const char *code);
@@ -38,15 +38,14 @@ static inline geohex_coordinate_ld_t _geohex_adjust_long_double_coordinate(const
 static inline geohex_coordinate_ld_t _geohex_location2coordinate(const geohex_location_t location) {
   geohex_coordinate_ld_t coordinate;
   coordinate.x = location.lng * GEOHEX3_HASH_BASE / 180.0L;
-  coordinate.y = GEOHEX3_HASH_BASE * logl(tanl((90.0L + location.lat) * (long double)M_PI / 360.0L)) / (long double)M_PI;
+  coordinate.y = GEOHEX3_HASH_BASE * logl(tanl((GEOHEX3_PI / 4.0L) + (location.lat * GEOHEX3_PI / 360.0L))) / GEOHEX3_PI;
   return coordinate;
 }
 
 static inline geohex_location_t _geohex_coordinate2location(const geohex_coordinate_ld_t coordinate) {
   geohex_location_t location;
   location.lng = 180.0L * coordinate.x / GEOHEX3_HASH_BASE;
-  location.lat = 180.0L * coordinate.y / GEOHEX3_HASH_BASE;
-  location.lat = 180.0L * (2.0L * atanl(expl(location.lat * (long double)M_PI / 180.0L)) - (long double)M_PI / 2.0L) / (long double)M_PI;
+  location.lat = 360.0L * atanl(expl(coordinate.y * GEOHEX3_PI / GEOHEX3_HASH_BASE)) / GEOHEX3_PI - 90.0L;
   return location;
 }
 
@@ -69,7 +68,7 @@ geohex_polygon_t geohex_get_hex_polygon (const geohex_t *geohex) {
   const long double h_x = coordinate.x;
   const long double h_y = coordinate.y;
 
-  const long double h_deg  = tanl(60.0L * (long double)M_PI / 180.0L);
+  const long double h_deg  = tanl(GEOHEX3_PI / 3.0L);
   const long double h_size = geohex->size;
 
   const long double center_lat     = geohex->location.lat;
@@ -139,15 +138,15 @@ geohex_coordinate_t geohex_get_coordinate_by_location(const geohex_location_t lo
   const long double h_x_q    = h_pos_x - h_x_0;
   const long double h_y_q    = h_pos_y - h_y_0;
 
-  long h_x = (long)roundl(h_pos_x);
-  long h_y = (long)roundl(h_pos_y);
-  if (h_y_q > -h_x_q + 1.0L) {
-    if ((h_y_q < 2.0L * h_x_q) && (h_y_q > 0.5L * h_x_q)) {
+  long h_x = lroundl(h_pos_x);
+  long h_y = lroundl(h_pos_y);
+  if (h_y_q + h_x_q > 1.0L) {
+    if ((h_y_q < (2.0L * h_x_q)) && ((2.0L * h_y_q) > h_x_q)) {
       h_x = (long)h_x_0 + 1L;
       h_y = (long)h_y_0 + 1L;
     }
-  } else if (h_y_q < -h_x_q + 1) {
-    if ((h_y_q > (2.0L * h_x_q) - 1.0L) && (h_y_q < (0.5L * h_x_q) + 0.5L)) {
+  } else if (h_y_q + h_x_q < 1.0L) {
+    if (((h_y_q + 1.0L) > (2.0L * h_x_q)) && ((2.0L * h_y_q) < (h_x_q + 1.0L))) {
       h_x = (long)h_x_0;
       h_y = (long)h_y_0;
     }
@@ -271,42 +270,50 @@ geohex_coordinate_t geohex_get_coordinate_by_code(const char *code) {
   return geohex_adjust_coordinate(geohex_coordinate(h_x, h_y), level);
 }
 
-geohex_t geohex_get_zone_by_coordinate(const geohex_coordinate_t coordinate, const geohex_level_t level) {
+static geohex_location_t geohex_get_location_by_coordinate(const geohex_coordinate_t coordinate, const geohex_level_t level) {
   const long double h_k    = _deg();
   const long double h_size = geohex_hexsize(level);
-
-  long double h_x = (long double)coordinate.x;
-  long double h_y = (long double)coordinate.y;
+  const long double x      = (long double)coordinate.x;
+  const long double y      = (long double)coordinate.y;
 
   const long double unit_x = 6.0L * h_size;
   const long double unit_y = 6.0L * h_size * h_k;
 
-  const long double       h_lat = (h_k * h_x * unit_x + h_y * unit_y) / 2.0L;
-  const long double       h_lon = (h_lat - h_y * unit_y) / h_k;
-  const geohex_location_t z_loc = _geohex_coordinate2location(_geohex_coordinate_ld(h_lon, h_lat));
+  const long double h_y = (h_k * x * unit_x + y * unit_y) / 2.0L;
+  const long double h_x = (h_y - y * unit_y) / h_k;
 
-  long double z_loc_x = z_loc.lng;
-  long double z_loc_y = z_loc.lat;
+  geohex_location_t z_loc = _geohex_coordinate2location(_geohex_coordinate_ld(h_x, h_y));
 
   const long max_hsteps = geohex_pow3(level + 2);
-  const long hsteps     = (long)floorl(fabsl(h_x - h_y));
+  const long hsteps     = labs(coordinate.x - coordinate.y);
   if (hsteps == max_hsteps) {
-    if (h_x > h_y) {
-      const long double tmp = h_x;
-      h_x = h_y;
-      h_y = tmp;
+    z_loc.lng = -180.0L;
+  }
+
+  return z_loc;
+}
+
+geohex_t geohex_get_zone_by_coordinate(const geohex_coordinate_t coordinate, const geohex_level_t level) {
+  const geohex_location_t z_loc = geohex_get_location_by_coordinate(coordinate, level);
+
+  long mod_x = coordinate.x;
+  long mod_y = coordinate.y;
+
+  const long max_hsteps = geohex_pow3(level + 2);
+  const long hsteps     = labs(mod_x - mod_y);
+  if (hsteps == max_hsteps) {
+    if (mod_x > mod_y) {
+      const long tmp = coordinate.x;
+      mod_x = mod_y;
+      mod_y = tmp;
     }
-    z_loc_x = -180.0L;
   }
 
   static char code3_x[GEOHEX3_DEC3_BUFSIZE/2];
   static char code3_y[GEOHEX3_DEC3_BUFSIZE/2];
-
-  long mod_x = h_x;
-  long mod_y = h_y;
   for (unsigned int i = 0; i <= level + 2; i++) {
     const long h_pow   = geohex_pow3(level + 2 - i);
-    const long h_pow_c = (long)ceill((long double)h_pow / 2.0L);
+    const long h_pow_c = h_pow / 2 + h_pow % 2;
 
     if (mod_x >= h_pow_c) {
       code3_x[i] = 2;
@@ -332,7 +339,7 @@ geohex_t geohex_get_zone_by_coordinate(const geohex_coordinate_t coordinate, con
       code3_y[i] = 1;
     }
 
-    if (i==2 && (z_loc_x == -180.0L || z_loc_x >= 0.0L)) {
+    if (i==2 && (z_loc.lng == -180.0L || z_loc.lng >= 0.0L)) {
       if (code3_x[0] == 2 && code3_y[0] == 1 && code3_x[1] == code3_y[1] && code3_x[2] == code3_y[2]) {
         code3_x[0] = 1;
         code3_y[0] = 2;
@@ -344,35 +351,35 @@ geohex_t geohex_get_zone_by_coordinate(const geohex_coordinate_t coordinate, con
     }
   }
 
-  static char h_code[GEOHEX3_DEC9_BUFSIZE];
-  for (unsigned int i = 0; i <= level + 2; i++) {
-    const static char chr[] = "0123456789";
+  static char h_code[GEOHEX3_DEC9_BUFSIZE - GEOHEX3_GLOBAL_CODE_BUFSIZE - 1];
+  static unsigned int global_code = 0;
+  {
+    const static unsigned int offset = GEOHEX3_GLOBAL_CODE_BUFSIZE - 1;
 
-    static char code3[GEOHEX3_DEC3_BUFSIZE];
-    code3[0] = chr[(unsigned)code3_x[i]];
-    code3[1] = chr[(unsigned)code3_y[i]];
-    code3[2] = '\0';
-
-    const int code9 = (int)strtol(code3, NULL, 3);
-    h_code[i] = chr[code9];
+    global_code = 0;
+    for (unsigned int i = 0, max = level + offset; i < max; i++) {
+      const unsigned int code9 = 3 * (unsigned int)code3_x[i] + (unsigned int)code3_y[i];
+      if (i < offset) {
+        global_code *= 10;
+        global_code += code9;
+      }
+      else {
+        const static char chr[] = "0123456789";
+        h_code[i-offset] = chr[code9];
+      }
+    }
+    h_code[level] = '\0';
   }
-  h_code[level + 3] = '\0';
-
-  static char global_code_buf[GEOHEX3_GLOBAL_CODE_BUFSIZE];
-  strncpy(global_code_buf, h_code, GEOHEX3_GLOBAL_CODE_BUFSIZE);
-  global_code_buf[GEOHEX3_GLOBAL_CODE_BUFSIZE-1] = '\0';
-  const int global_code = atoi(global_code_buf);
 
   geohex_t geohex = {
-    .location   = geohex_location(z_loc_y, z_loc_x),
+    .location   = geohex_location(z_loc.lat, z_loc.lng),
     .coordinate = coordinate,
     .level      = level,
-    .size       = h_size
+    .size       = geohex_hexsize(level)
   };
   geohex.code[0] = GEOHEX3_HASH_KEY[(int)floor(global_code/30)];
   geohex.code[1] = GEOHEX3_HASH_KEY[global_code%30];
-  strncpy(geohex.code + 2, h_code + 3, GEOHEX3_MAX_LEVEL);
-  geohex.code[strlen(h_code)] = '\0';
+  strncpy(geohex.code + 2, h_code, level+1);
   return geohex;
 }
 
